/******************************************************************************/
// Zeitschaltuhr Funk
// Schaltuhr fr Funksteckdosen
//
// loetmeister.de - 2016, v0.9
/******************************************************************************/
//
// http://www.loetmeister.de/Elektronik/zeitschaltuhr-funk.htm
//
// Prozessor: PIC16F1827
// Prozessortakt:  8 MHz intern
// Eingangsspannung: 5V

// PIC16F1827 Device ID Revision = 3: Workarounds as per errata DS80485H document

/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/
#if defined(__XC)
    #include <xc.h>         /* XC8 General Include File */
#elif defined(HI_TECH_C)
    #include <htc.h>        /* HiTech General Include File */
#endif

#include <stdint.h>        /* For uint8_t definition */
#include <stdbool.h>       /* For true/false definition */

#include "system.h"        /* System funct/params, like osc/peripheral config */
#include "user.h"          /* User funct/params, such as InitApp */

#include "eeprom/eeprom_custom_routines.h"  /* Angepasste Funnktionen */
#include "lcd_hd44780_pic16/lcd_hd44780_pic16.h"

/******************************************************************************/
/* User Global Variable Declaration                                           */
/******************************************************************************/

// Status and IO
bit UpdateDisplay;  // update LCD
bit DCFneueDaten;      // Neuer DCF77 Datensatz bereit
bit DisplayON;  // display is on
bit LowBatt;
bit IOResetDone;
/* status of output channels */
volatile unsigned char IORelayCurrentState, IORelayNewState;

// timer
unsigned long SystemTicker = 0;
//unsigned char GelaufenInSekunde = 0;
unsigned char GelaufenInMinute = 0;
unsigned char GelaufenAmTag = 0;
bit SystemTickerZwanzigMs;

// RF Transmit
unsigned char RFTXRepeatCounter = RFTX_NUMBER_OF_SAMPLES;
unsigned char RFTXSendBuffer[16];          // holds the entire frame to be send (filled before send rountine is started)
unsigned char RFTXSendBufferIndex = 0;     // current index no. of the byte we are sending from the buffer array
unsigned char RFTXSendbyte;                // the byte we are sending
unsigned char RFTXSendbyteActiveBitPos;    // active bit position of the byte we are sending
unsigned char RFTXSendbyteActiveBit = 0;   // current bit on the "wire"
unsigned char RFTXTimerPR;                 // timer period register (Timer2/4/6 can trigger an interrupt on TMRx-to-PRx match)

// TX code: 0 = 0b01000100 (68), 1 = 0b01110111 (119), F = 0b01000111 (71)
const unsigned char rftx_sync[4] = {HS2260A_CodeLogical_One,HS2260A_CodeLogical_F,HS2260A_CodeLogical_F,HS2260A_CodeLogical_F}; /* init sequence */
const unsigned char rftx_channel[RFTX_NUMBER_OF_RECEIVERS][4] = {
    {HS2260A_CodeLogical_One,HS2260A_CodeLogical_F,HS2260A_CodeLogical_One,HS2260A_CodeLogical_Zero}, /* first channel */
    {HS2260A_CodeLogical_F,HS2260A_CodeLogical_One,HS2260A_CodeLogical_F,HS2260A_CodeLogical_Zero}, /* second channel */
    {HS2260A_CodeLogical_F,HS2260A_CodeLogical_F,HS2260A_CodeLogical_One,HS2260A_CodeLogical_Zero} /* third channel */
}; // channel code
const unsigned char rftx_command[2][4] = {
    {HS2260A_CodeLogical_Zero,HS2260A_CodeLogical_Zero,HS2260A_CodeLogical_Zero,HS2260A_CodeLogical_One}, /* off */
    {HS2260A_CodeLogical_Zero,HS2260A_CodeLogical_Zero,HS2260A_CodeLogical_One,HS2260A_CodeLogical_Zero}  /* on */
}; // send on/off command
const unsigned char rftx_end[4] = {64,0,0,0}; /* end sequence of each frame */

unsigned char WatchDogSleepCounter = 0;
unsigned char PowerOffTimer = POWER_OFF_DELAY;
//unsigned int ADCBatteryReading;

// Uhr
volatile struct
{
    unsigned char Wochentag;         // Wochentag (1-7)
#ifndef REDUCED_CODE_SIZE
    unsigned char Tag;               // Tag (1-31)
    unsigned char Monat;             // Monat (1-12)
#endif
    unsigned char Stunde;            // Aktuelle Stunde (0-23)
    unsigned char Minute;            // Minuten (0-59)
    unsigned char Sekunde;           // Minuten (0-59)
} Kalender;

#ifndef REDUCED_CODE_SIZE
const char Monatslaenge[12]={31,28,31,30,31,30,31,31,30,31,30,31};
#endif
volatile unsigned char BitZeigerIndex, BitZeigerShift;
volatile unsigned char BitZeiger = 0;             // Zeiger auf das aktuelle Bit im Puffer
volatile unsigned char Bitlaenge = 0;             // Zhler fr die Dauer des aktuellen Bits
bit BitStart;                            // Luft gerade eine Zeitmessung?
volatile unsigned long BitPuffer[2];     // Speicherung des Rohdatensatz
volatile unsigned char DCFMinutensyncTimer = 170; // Zeitmessung 1,7 Sec
volatile unsigned char StoerungsfreieZeit = 0;
volatile unsigned char DCFsyncMinuteAlt = 0, DCFsyncStundeAlt = 0;   // vorherige Minute Speichern, zum vergleich und Erkennung zweier DCF signale in Folge
volatile signed char ZeitOffsetMinuten = 0;
bit TimeOutOfSync;

union tagRelay
{
    struct tagTimerRbyte                 // EEPROM data layout: Timer relay operation and assignment
    {   // LSB on top
//        unsigned RelayOutput:4;     // New output state
//        unsigned RelayOperate:4;    // Neet to operate the specific relay? (change state or leave alone?)
        unsigned RelayOne:2;
        unsigned RelayTwo:2;
        unsigned RelayThree:2;
        unsigned RelayFour:2;
//        bit: Relay output - set or clear?
//        bit: Operate Relay?
    } bits;
    unsigned char byte:8;       // entire byte
} TimerR, AktuelleSchaltzeitR[SCHALTZEITEN_PRO_TAG_MAX];

union tagMinuteAndRandom
{
    struct tagTimerMbyte                 // EEPROM data layout: Timer minutes and Random flag
    {   // LSB on top
        unsigned MINUTE:7;              // Minutes (would actually need only 6 bits)
        unsigned Random:1;              // Random enabled? (delay relay operation by random value)
    } bits;
    unsigned char byte:8;       // entire byte
} TimerM, AktuelleSchaltzeitM[SCHALTZEITEN_PRO_TAG_MAX];

volatile unsigned char AktuelleSchaltzeitStunde[SCHALTZEITEN_PRO_TAG_MAX];  // x Schaltzeiten pro Wochentag mglich
volatile unsigned char AnzahlAktuelleSchaltzeiten = 0;
volatile unsigned char TimerRandomMinuteMax;

// buttons
unsigned int ButtonCompare[6] = {933,840,733,627,525,423};  // {934,842,733,627,525,423}
//unsigned char cmp[6] = {93,83,72,63,52,42};
unsigned char ButtonDebounceCounter = BUTTON_DEBOUNCE_COUNT;
volatile unsigned char ButtonON = NO_BUTTON;
unsigned char ButtonON_last = NO_BUTTON;

// menu related
volatile unsigned char MenuIdleTimer = MENU_IDLE_TIMER/3;    // timeout for home screen (use 1/3 of the time for initial start)
bit MenuChangeValue;
unsigned char MenuCurrentWeekday;
volatile unsigned char MenuItem = 0, MenuLevel = 0;
volatile unsigned char *Buff_NewValue;
volatile unsigned char Buff_NewValue_EEPROMAddr;
volatile unsigned char Timer_EEPROMAddr_Wday = EE_ADDR_TIMER_START;
unsigned char Timer_Weekday_byte;
unsigned char Timer_hour_byte;
unsigned char Timer_Minute;
bit MenuChangeTimeHour;
bit EEPROMUpdateFlag;    // signal main loop to reload EEPROM values

typedef struct MENU {
    unsigned char level;
    const char *text;
//    void (*f_pointer)(void); // funktion die aufgerufen werden soll // oder: void (*f_pointer) ( unsigned char wert);
    void (*tastfun[3])(void);   /*tastenfunktionen*/
    void (*m_pointer)(void);    /*menfunktion, falls vorhanden*/
} tMenu;

// LCD static text
const char LCD_Welcome[] = "Hallo!";
const char LCD_Space2[] = "  ";
//const char LCD_RelayOne[] = "Rel1";
//const char LCD_RelayTwo[] = "Rel2";
const char LCD_RelayOne[] = "A";
const char LCD_RelayTwo[] = "B";
const char LCD_RelayThree[] = "C";
const char LCD_RelayFour[] = "D";

const char LCD_ON[]  = "Ein";
const char LCD_OFF[] = "Aus";
const char LCD_Weekdays[] = " MDMDFSS";
const char menu_string_timer[] = "S.Zeit";
const char menu_string_rnd[] = "Zuf";
const char menu_string_memory[] = "Nr.";
const char menu_string_day[] = "Tag";
//const char menu_string_time[] = "Zeit";
const char menu_string_back[] = "Zur\365ck"; // =\365, =\357, =\341
const char menu_string_hour[] = "Std";
const char menu_string_minute[] = "Min";
const char menu_string_clear_timer[] = "Reset";
const char menu_string_confirm[] = "OK?";
const char menu_string_battery[] = "Batt:";
const char menu_string_battery_low[] = "ko";
const char menu_string_battery_ok[] = "ok";

#define INFO_DISPLAY_LINES 3    // no cursor for the first n lines

#define LOW_BAT_MENU_LEVEL 1    // on which menu level is the battery status entry?
#define LOW_BAT_MENU_ITEM 28    // index number in menu array

/* Komplette Menstruktur */
const tMenu MainMenu[]= {
/*  level, static txt, {UP, DOWN, OK tastfunc}, display func */
{0, LCD_Welcome, {dummyfunc, dummyfunc, dummyfunc}, dummyfunc},
//{0, LCD_Welcome2, {dummyfunc, dummyfunc, dummyfunc}, dummyfunc},    // optional - when enabled, need to update MenuItem = 2
    {1, 0, {dummyfunc, menu_curdown, dummyfunc}, print_time_status},
    {1, 0, {menu_curup, menu_curdown, dummyfunc}, print_relais_status},
    {1, 0, {menu_curup, menu_curdown, dummyfunc}, print_relais_statusTwo}, /* no selection possible, just info: INFO_DISPLAY_LINES */
    
    // Erstellen und Bearbeiten der Schaltzeiten
    {1, menu_string_timer, {menu_curup, menu_curdown, menu_enter_timer}, dummyfunc},
        {2, menu_string_memory, {menu_curup, menu_curdown, menu_enter_select_timer_memory}, print_timer_memory_address},
            {3, 0, {menu_timer_addr_up, menu_timer_addr_down, menu_exit_select_timer_memory}, print_timer_memory_address},
            {3, 0, {dummyfunc, dummyfunc, dummyfunc}, print_timer_weekday},
        {2, menu_string_day, {menu_curup, menu_curdown, menu_enter_change_timer_weekday}, dummyfunc},
            {3, 0, {menu_relais_weekday_toggle, menu_relais_weekday_toggle, menu_exit_change_relay}, Print_SelectWeekday},
            {3, LCD_Weekdays, {dummyfunc, dummyfunc, dummyfunc}, dummyfunc},
        {2, 0, {menu_curup, menu_curdown, menu_enter_timer_change_time_or_relay}, print_timer_new_time},
            {3, 0, {menu_timer_change_time_up, menu_timer_change_time_down, menu_exit_timer_change_time}, print_timer_new_time},
            {3, 0, {dummyfunc, dummyfunc, dummyfunc}, print_timer_new_time_hhmm}, // depending on selection, show hour or minute in second line
        {2, LCD_RelayOne, {menu_curup, menu_curdown, menu_enter_timer_change_time_or_relay}, print_timer_relay_status_one},
            {3, LCD_RelayOne, {menu_timer_change_relay_one, menu_timer_change_relay_one, menu_exit_timer_change_relay}, print_timer_relay_status_one},
        {2, LCD_RelayTwo, {menu_curup, menu_curdown, menu_enter_timer_change_time_or_relay}, print_timer_relay_status_two},
            {3, LCD_RelayTwo, {menu_timer_change_relay_two, menu_timer_change_relay_two, menu_exit_timer_change_relay}, print_timer_relay_status_two},
        {2, LCD_RelayThree, {menu_curup, menu_curdown, menu_enter_timer_change_time_or_relay}, print_timer_relay_status_three},
            {3, LCD_RelayThree, {menu_timer_change_relay_three, menu_timer_change_relay_three, menu_exit_timer_change_relay}, print_timer_relay_status_three},
//        {2, LCD_RelayFour, {menu_curup, menu_curdown, menu_enter_timer_change_time_or_relay}, print_timer_relay_status_two},
//            {3, LCD_RelayFour, {menu_timer_change_relay_four, menu_timer_change_relay_four, menu_exit_timer_change_relay}, print_timer_relay_status_four},
        {2, menu_string_rnd, {menu_curup, menu_curdown, menu_enter_timer_change_time_or_relay}, print_timer_random_enabled},
            {3, menu_string_rnd, {menu_timer_change_random_enabled, menu_timer_change_random_enabled, menu_exit_timer_change_relay}, print_timer_random_enabled},
        {2, menu_string_clear_timer, {menu_curup, menu_curdown, menu_entersub}, dummyfunc},
            {3, menu_string_confirm, {menu_back, menu_back, menu_exit_timer_clearall}, dummyfunc},
        {2, menu_string_back, {menu_curup, menu_curdown, menu_back}, dummyfunc},
    
    // ndern der maximal zulssigen Zufallszeit
    {1, menu_string_rnd, {menu_curup, menu_curdown, menu_enter_change_TimerRandomMax}, dummyfunc},
        {2, 0, {menu_enter_TimerRandomMax_up, menu_enter_TimerRandomMax_down, menu_exit_eeprom_save_value}, print_changed_value},
        {2, 0, {dummyfunc, dummyfunc, dummyfunc}, print_timer_new_time_hhmm},
    
    // Anzeigen des Batterie status (LOW_BAT_MENU_ITEM & LOW_BAT_MENU_LEVEL setzen!!)
    {1, menu_string_battery, {menu_curup, dummyfunc, dummyfunc}, print_battery_status},  /*last entry on main level!! */

    {1, 0, {menu_curup, dummyfunc, dummyfunc}, print_debug_status}
};


/******************************************************************************/
/* Main Program                                                               */
/******************************************************************************/
void main(void)
{
    unsigned char i,r,a,ChannelNum = 0;
    
    /* Configure the oscillator for the device */
//    ConfigureOscillator(); // no function to save program space
//    OSCCONbits.IRCF = 0b1110; // =8 MHz or 32 MHz HF (depends on PLL 4x setting)
//    OSCCONbits.IRCF = 0b1111; // =16 MHz
    set_cpu_seed_low(); // start with slow speed. For RF Module transmission, speed needs to be increased
    
    /* Initialize I/O and Peripherals for application */
    InitApp();
    
    UpdateDisplay = true;
    
    read_eeprom_values();
    
    Init_timer4();  /* 10 ms system timer */
    
    Enable_interrupts();
    
    while(1)
    {
      /* Die Zeit erst auswerten wenn ein kompletter Datensatz empfangen wurde */
        if (DCFneueDaten)
        {
            DekodiereZeit();
        }
        
        
        if (ZeitOffsetMinuten > 1)  // Uhr vorstellen
        {
            TMR4IF = false;
            Kalender.Sekunde = 59;
            ZeitOffsetMinuten--;
            TMR4IF = true;
        }
        else
            TimeOutOfSync = false;
        
        
      /* Tglich, oder wenn sich der Tag nach dem DCF sync ndert */
//        if (Kalender.Tag != GelaufenAmTag || EEPROMUpdateFlag)
        if (Kalender.Wochentag != GelaufenAmTag || EEPROMUpdateFlag)
        {
//            GelaufenAmTag = Kalender.Tag;
            GelaufenAmTag = Kalender.Wochentag;
            read_eeprom_values();
        }
        
                
        if (StoerungsfreieZeit > 0)  // Uhr ok?
        {
        /* nur einmal pro Minute */
            if (Kalender.Minute != GelaufenInMinute)
            {
                GelaufenInMinute = Kalender.Minute;
                
                /* Schaltzeiten dieses Tages prfen und Ausfhren */
                if (AnzahlAktuelleSchaltzeiten > 0)
                {
                    for (i=0; i < AnzahlAktuelleSchaltzeiten; i++)
                    {
                        if ((AktuelleSchaltzeitStunde[i] == Kalender.Stunde) && (AktuelleSchaltzeitM[i].bits.MINUTE == GelaufenInMinute))
                        {
                            a = 0;
                            for (r=0;r<((RFTX_NUMBER_OF_RECEIVERS*2)-1);r+=2)   // loop only through the configured receivers
                            {
                                if (AktuelleSchaltzeitR[i].byte & (1 << r +1)) // assignment bit set? If not, do not change relay state
                                {
                                    if (AktuelleSchaltzeitR[i].byte & (1 << r))
                                        IORelayNewState |= (1 << a);    // turn on
                                    else
                                        IORelayNewState &= ~(1 << a);    // turn off
                                }
                                a++;
                            }
                        }
                    }
                } // prfe schaltzeiten
            }
        }
        else if (StoerungsfreieZeit == 0 && !IOResetDone)     // DCF Signal weg - alles aus, aber nur einmal Senden
        {
//            IORelayCurrentState = 0b00001111;     // nderung erzwingen (4 Empfnger)
            IORelayCurrentState = 0b00000111;     // nderung erzwingen (3 Empfnger)
            IORelayNewState = 0;      // Ziel: Alles aus
            IOResetDone = true;
        }
        
        
        if (!RFTX_Active)   // wait if still tranmitting data (cannot send more than one signal, ADC is not working with high CPU speed anyways)
        {
            /* every 20ms */
            if (SystemTickerZwanzigMs)
            {
                SystemTickerZwanzigMs = false;
                
                ReadButtons();      // 20ms * 8 ADC readings == 160 ms debounce

                /* button 0: "up", 1: "down", 2: "ok" */
                if (ButtonON <= 2)       /* taste holen und entsprechende aktion auslsen */
                {
                    MainMenu[MenuItem].tastfun[ButtonON]();
                    UpdateDisplay = true;
                    PowerOffTimer = POWER_OFF_DELAY;
                }
                /* button 3: "A", 4: "B", 5: "C" */
                else if (ButtonON <= 5)    // operate IO directly
                {
                    IORelayNewState ^= (1 << (ButtonON -3));    // toggle channel
                }
                ButtonON = NO_BUTTON;
                
                /* send new data to display when needed */
                if (UpdateDisplay)
                {
                    if (!DisplayON) // power enabled and display initialized?
                    {
                        IO_DCDC_Module_On();
                        DisplayON = true;
                        __delay_ms(40);    // allow voltage to stabilize
                        LCDInit(LS_NONE);   // init LCD, use no cursor
                        GetBatteryStatus();
                        
                        if (LowBatt) // jump to battery status menu entry, as warning if battery is low
                        {
                            MenuLevel = LOW_BAT_MENU_LEVEL;
                            MenuItem = LOW_BAT_MENU_ITEM;
                        }
                    }
                    LCD_display();
                }
                
                
                /* IO need to change? */
                if((IORelayNewState != IORelayCurrentState))// && !TimeOutOfSync)
                {
//                    if (!PowerON) // power enabled? If not, enable and wait 20ms by system ticker
                    if (IO_DCDC_Module_GetValue()) // power enabled? If not, enable and wait 20ms by system ticker
                    {
                        IO_DCDC_Module_On();
                    }
                    else   // operate only one channel at a time
                    {
                        if (IORelayNewState & (1 << ChannelNum))    // TODO? - sending command, not checking if actual value was already ok
                            RFTX_Send(ChannelNum,ON);
                        else
                            RFTX_Send(ChannelNum,OFF);

                        ChannelNum++;
                        if (ChannelNum >= RFTX_NUMBER_OF_RECEIVERS)
                            ChannelNum = 0;
                    }
                }
                else if (PowerOffTimer == 0)  // nothing to send and PowerOffTimer counted down: turn power off
                {
                    IO_DCDC_Module_Off();
                    LATA = 0x00;    // need to clear ports connected to LCD, init fails if not done
                    DisplayON = false;
                }
            }
            
            Enable_Watchdog();
            SLEEP();    // sleep for WDT timeout period
            Disable_Watchdog();  // Disable Watchdog after wake-up (pre-loaded instruction)
            
            WatchDogSleepCounter++;
            if (WatchDogSleepCounter >= 4) // ==5
            {
                /* reset counter and trigger system ticker interrupt - the timer did not run during sleep! */
                TMR4IF = false;
                WatchDogSleepCounter = 0;
                TMR4IF = true;
            }
        }
    } // while 1

} // void main


/******************************************************************************/
/*  EEPROM Werte beich Start/Reset auslesen                                   */
/******************************************************************************/
void read_eeprom_values(void)
{
    /* eeprom_read(0x00)   // from 0x00 to 0xFF */

    unsigned char eeaddr;
    unsigned char WochentageBits, SchaltzeitMinuteNeu;
    unsigned int Seed, OldSeed = 0;
    unsigned char SeedCount, RNDIterations;
    unsigned char TimerRandomMinute;
    
    // Maximaler Wert fr zufllige Schaltzeitenverschiebung auslesen
    TimerRandomMinuteMax = eeprom_read(EE_ADDR_TIMER_RND_MAX);   
    
/*  Schaltzeiten fr den aktuellen Tag aus dem EEPROM lesen.                  */
    // Gltiger Wochentag 1...7
    if (Kalender.Wochentag != 0)
    {
        AnzahlAktuelleSchaltzeiten = 0;
        
        // alle EEPROM Zellen mit Wochentaginformationen fr Schaltzeiten lesen
        for (eeaddr=EE_ADDR_TIMER_START; eeaddr < EE_ADDR_TIMER_STARTa_H; eeaddr++)
        {
            WochentageBits = eeprom_read(eeaddr);
            
            /* Ist das bit fr den jeweiligen Tag gesetzt? */
            if (WochentageBits & (1 << Kalender.Wochentag))
            {
                AktuelleSchaltzeitStunde[AnzahlAktuelleSchaltzeiten] = eeprom_read(eeaddr+EE_ADDR_TIMER_STARTo_H); // + hour offset
                AktuelleSchaltzeitM[AnzahlAktuelleSchaltzeiten].byte = eeprom_read(eeaddr+EE_ADDR_TIMER_STARTo_M); // + minute offset
                AktuelleSchaltzeitR[AnzahlAktuelleSchaltzeiten].byte = eeprom_read(eeaddr+EE_ADDR_TIMER_STARTo_R); // + relay offset
                
                if (AktuelleSchaltzeitM[AnzahlAktuelleSchaltzeiten].bits.Random)    // Zufall aktiviert?
                {
                    RNDIterations = 48;     // maximum iterations to find a random value within the limit
                    
                    do
                    {
                        if (OldSeed == 0)
                        {
                            for (SeedCount=0; SeedCount<15; SeedCount++)
                                OldSeed += OldSeed + ZufallsBit();   //init
                        }
                        
                        Seed = 181*(OldSeed) + 359;// 8 bit random number=high(SEED)
                        OldSeed = Seed;
//                        Zufall = Seed >>11; // only use 5 bits
//                        TimerRandomMinute = Zufall + ZufallsBit();
                        TimerRandomMinute = (Seed >>11) + ZufallsBit(); // only use 5 bits and add another random bit
                        
                        RNDIterations--;    // don't stay in this loop forever
                        if (RNDIterations == 0)
                        {
                            TimerRandomMinute = 0;  // no random value found - keep original time
//                            break;
                        }
                    }
                    while (TimerRandomMinute > TimerRandomMinuteMax);
                    
                    SchaltzeitMinuteNeu = AktuelleSchaltzeitM[AnzahlAktuelleSchaltzeiten].bits.MINUTE + TimerRandomMinute;
                    if ((SchaltzeitMinuteNeu >= 60) && (AktuelleSchaltzeitStunde[AnzahlAktuelleSchaltzeiten] <= 22))    // kein Tag wechsel mglich...
                    {
                        SchaltzeitMinuteNeu -= 60;
//                        if (AktuelleSchaltzeitStunde[AnzahlAktuelleSchaltzeiten] <= 22)
                        AktuelleSchaltzeitStunde[AnzahlAktuelleSchaltzeiten] += 1;
//                        else
//                        {
//                            AktuelleSchaltzeitStunde[SCHALTZEITEN_PRO_TAG_MAX + AnzahlSchaltzeitenUebertrag] = 0;
//                            AktuelleSchaltzeitM[SCHALTZEITEN_PRO_TAG_MAX + AnzahlSchaltzeitenUebertrag].bits.MINUTE = SchaltzeitMinuteNeu;
//                            AktuelleSchaltzeitR[SCHALTZEITEN_PRO_TAG_MAX + AnzahlSchaltzeitenUebertrag].byte = AktuelleSchaltzeitR[AnzahlAktuelleSchaltzeiten].byte;
//                        }
                    }
                    if (SchaltzeitMinuteNeu < 60)
                        AktuelleSchaltzeitM[AnzahlAktuelleSchaltzeiten].bits.MINUTE = SchaltzeitMinuteNeu;
                }
                AnzahlAktuelleSchaltzeiten++;
            }
            if (AnzahlAktuelleSchaltzeiten >= SCHALTZEITEN_PRO_TAG_MAX)   // we only use the _MAX defined values, found from start address
                break;   
        }
    }
    
    EEPROMUpdateFlag = false;
}

/******************************************************************************/
/* Ein zuflliges bit, mit dem rauschen einer LED, erzeugen                   */
/******************************************************************************/
unsigned char ZufallsBit(void)
{
    ADON = true;   // ADC einschalten
    ADC_Channel_Select_Rnd();    // Select input channel for LED
    adc_read();
    
    return ADRESL & 0b00000001;
}

/******************************************************************************/
/*  Schaltzeiten fr den aktuellen Tag aus dem EEPROM lesen.                  */
/******************************************************************************/
//void TaeglicheSchaltzeitenLesen(void)
//{
//    // Gltiger Wochentag 1...7
//    if (Kalender.Wochentag != 0)
//    {
//        unsigned char i, WochentageBits;
//
//        AnzahlAktuelleSchaltzeiten = 0;   // TODO - Anzahl der Tglichen Schaltzeiten im EEPROM ablegen?
//
//        // alle EEPROM Zellen mit Wochentag informationen fr Schaltzeiten lesen
//        for (i=EE_ADDR_TIMER_START; i < EE_ADDR_TIMER_END-1; i+=EE_INC_ADDR_TIMER)
//        {
//            WochentageBits = eeprom_read(i);
//            if ((WochentageBits & (1 << Kalender.Wochentag)) && (AnzahlAktuelleSchaltzeiten < 7))   // Ist das bit fr den jeweiligen Tag gesetzt?
//            {
//                AktuelleSchaltzeitStunde[AnzahlAktuelleSchaltzeiten] = eeprom_read(i+1);
//                AktuelleSchaltzeitMR[AnzahlAktuelleSchaltzeiten].byte = eeprom_read(i+2);
//
//                AnzahlAktuelleSchaltzeiten++;
//            }
//        }
//    }
//}

/******************************************************************************/
/*  Temperaturmessung mit internem Sensor (on chip)                           */
/******************************************************************************/
//void MesseTemperaturIntern(void)
//{
//    ADON = true;   // ADC einschalten
//    FVRCONbits.TSEN = true; // enable temperature sensor
//    ADCON0bits.CHS = 0b11101;    // Select temperature sensor channel
//
//    __delay_us(2);
//    adc_read();
//
//    ADCout.ad_hi_low[0] = ADRESL;
//    ADCout.ad_hi_low[1] = ADRESH;
//
//    ADON = false;   // ADC ausschalten
//
////    ADCout.ADRES -= TemperaturKalibierung;
//}

/******************************************************************************/
/* Auslesen und Auswerten der Batterie Spannung (ber Spannungsreferenz)      */
/******************************************************************************/
void GetBatteryStatus(void)
{
//    unsigned int ADCBatteryReading;
    
    /* enable Fixed Voltage Reference Peripheral */
    FVRCON = 0b10000010;   // FVREN = 1, FVRRDY 0, Reserved 00, CDAFVR = 00, ADFVR = 01: ADC Fixed Voltage Reference Peripheral output is 2x (2.048V)
    while (!FVRRDY)     // wait until ready
            ;
    ADCON1bits.ADPREF = 0b11;   // VREF+ is connected to internal Fixed Voltage Reference (FVR) module
    
    ADON = true;   // ADC einschalten
    ADC_Channel_Select_Batt();    // Select input channel for battery voltage
    adc_read();
    
    FVRCON = 0;     // turn Fixed Voltage Reference Peripheral off
    ADCON1bits.ADPREF = 0;  // VREF+ is connected to VDD again
    
//    ADCBatteryReading = ADRESL + (unsigned int)(ADRESH << 8);
    
    LowBatt = false;
//    if (ADCBatteryReading < LOW_BATTERY_LEVEL)
    if ((ADRESL + (unsigned int)(ADRESH << 8)) < LOW_BATTERY_LEVEL)
        LowBatt = true;  // keep this warning until bateries where replaced
}

/******************************************************************************/
/* Einlesen der Taster (Spannungsteiler am ADC Eingang)                      */
/******************************************************************************/
void ReadButtons(void)
{
    unsigned int ADCButtonReading;
    unsigned char i;
    
    ADON = true;   // ADC einschalten
    ADC_Channel_Select_Buttons();    // Select input channel for buttons
    adc_read();
    
    ADCButtonReading = ADRESL + (unsigned int)(ADRESH << 8);
    
    if (ADCButtonReading > 100) // Any button pressed? (reached min. voltage level?)
    {
        for (i=0;i<6;i++)
        {
            if ((ADCButtonReading < ButtonCompare[i]+12) && (ADCButtonReading > ButtonCompare[i]-12))
            {
                if (ButtonON_last == i)
                {
                    if (ButtonDebounceCounter > 0)
                        ButtonDebounceCounter--;
                    else
                    {
                        ButtonON = i;
                        ButtonDebounceCounter = BUTTON_DEBOUNCE_COUNT*3;    // additional repeat delay (when button remains pressed)
                    }
                }
                else
                {
                    ButtonON_last = i;
                    ButtonDebounceCounter = BUTTON_DEBOUNCE_COUNT;
                }
                
                MenuIdleTimer = MENU_IDLE_TIMER;
            }
        }
    }
    else
    {
        ButtonON_last = NO_BUTTON;
        ButtonDebounceCounter = BUTTON_DEBOUNCE_COUNT;
    }
}

/******************************************************************************/
/* Senden ber das Funkmodul. Kanal A,B,C=0,1,2; Befehl: An=1, Aus=0          */
/******************************************************************************/
void RFTX_Send(unsigned char channel, unsigned char command)
{
    if (!RFTX_Active && channel <= (RFTX_NUMBER_OF_RECEIVERS-1))// && command <= 1) // valid range....
    {
        unsigned char c;

        // load transmit buffer
        for (c=0; c<4; c++)
        {
            RFTXSendBuffer[c] = rftx_sync[c];
            RFTXSendBuffer[c+4] = rftx_channel[channel][c];
            RFTXSendBuffer[c+8] = rftx_command[command][c];
            RFTXSendBuffer[c+12] = rftx_end[c];
        }
        
        // set new state flags
        if (command == ON)
            IORelayCurrentState |= (1 << channel);
        else
            IORelayCurrentState &= ~(1 << channel);

        // prepare final variables and turn on fast CPU speed
        RFTXSendbyteActiveBitPos = 7;
        set_cpu_seed_high();
        TMR6IF = false;        // clear interrupt
        RFTX_Active = true;    // enable Timer6 interrupt - which handles the actual transmission
    }
}

/******************************************************************************/
/* Ein- oder Ausschalten des DC-DC Wandlers.                                  */
/******************************************************************************/
//void DCDC_Power_On_Off(void)
//{
//    /* operate DC-DC module for LCD & TX */
//}

/******************************************************************************/
/* berprft auf gerade Paritt                                               */
/******************************************************************************/
char ParitaetsCheck(char wert)
 {
    char Einsen = 0;
    char i;

    // Anzahl der 1-Bits zhlen
    for (i = 0; i <8; i++)
    {
        if (wert & (1 << i))
            Einsen++;
    }
    
    return (!(Einsen % 2));   // Anzahl muss gerade sein, also Modulo 2 == 0, sonst Fehler
 }

/******************************************************************************/
/* Erzeugt Datum und Uhrzeit                                                  */
/******************************************************************************/
void DekodiereZeit(void)
{
    struct
    {
        unsigned char Wochentag;         // Wochentag (1-7)
        unsigned char Tag;               // Tag (1-31)
#ifndef REDUCED_CODE_SIZE
        unsigned char Monat;             // Monat (1-12)
#endif
        unsigned char Stunde;            // Aktuelle Stunde (0-23)
        unsigned char Minute;            // Minuten (0-59)
    } TimerBCD;
	
	unsigned char DCFsyncMinute, DCFsyncStunde;

    //---------------------------
    // Datensatz zerlegen
    //---------------------------

    // Minuten erzeugen
    TimerBCD.Minute = (char) ((BitPuffer[0] & MaskeMinuten) >> ShiftMinuten);

    // Stunden erzeugen (Verteilt ber 2 Longs)
    TimerBCD.Stunde =  (char) ((BitPuffer[0] & MaskeStunden1) >> ShiftStunden1);
    TimerBCD.Stunde |= (char) ((BitPuffer[1] & MaskeStunden2) << ShiftStunden2);
    // Parittsbit nach aussen verschieben
    if (TimerBCD.Stunde & 0b01000000) TimerBCD.Stunde |= 0b10000000;
    TimerBCD.Stunde &= ~0b01000000;

    // Tag erzeugen
    TimerBCD.Tag = (char) ((BitPuffer[1] & MaskeTag) >> ShiftTag);

    // Wochentag erzeugen
    TimerBCD.Wochentag = (char) ((BitPuffer[1] & MaskeWochentag) >> ShiftWochentag);

#ifndef REDUCED_CODE_SIZE
    // Monat erzeugen
    TimerBCD.Monat = (char) ((BitPuffer[1] & MaskeMonat) >> ShiftMonat);
#endif

#ifndef REDUCED_CODE_SIZE
    if ((TimerBCD.Tag != 0) && (TimerBCD.Wochentag != 0) && (TimerBCD.Monat != 0) && (ParitaetsCheck(TimerBCD.Stunde)) && (ParitaetsCheck(TimerBCD.Minute)))
#else
    if ((TimerBCD.Tag != 0) && (TimerBCD.Wochentag != 0) && ParitaetsCheck(TimerBCD.Stunde) && ParitaetsCheck(TimerBCD.Minute))    
#endif
    {
        // Minuten dezimal umbauen
        DCFsyncMinute =    TimerBCD.Minute & 0b00001111;
        DCFsyncMinute += ((TimerBCD.Minute & 0b01110000) >> 4) * 10;
        
        // Stunden dezimal umbauen
        DCFsyncStunde =    TimerBCD.Stunde & 0b00001111;
        DCFsyncStunde += ((TimerBCD.Stunde & 0b00110000) >> 4) * 10;       
        
        // Es mssen zwei korrekte Datenstze in Folge empfangen werden (ignoriere Wechel zur neuen Stunde)
        if ((DCFsyncMinuteAlt +1) == DCFsyncMinute && DCFsyncStundeAlt == DCFsyncStunde)
        {
            //--------------------------------------
            // Daten dezimal umbauen und bernehmen
            //--------------------------------------
            
            ZeitOffsetMinuten = DCFsyncMinute - Kalender.Minute;    // Berechung nur innerhalb der selben Stunde!!
            
            if ((StoerungsfreieZeit > 0) && (ZeitOffsetMinuten > 1) && (ZeitOffsetMinuten < 24))
                TimeOutOfSync = true;   // Uhr war im sync, differenz Zeit zu DCF Zeit > 1 Minute, dann offset berechnen und langsame Korrektur ausfhren
            else // Kompletter neu sync, wenn offset 1 Minute oder Uhr in Strung
            {
                TimeOutOfSync = false;

                Kalender.Sekunde = 1;

//                // Minuten
        //        Kalender.Minute =    TimerBCD.Minute & 0b00001111;
        //        Kalender.Minute += ((TimerBCD.Minute & 0b01110000) >> 4) * 10;
                Kalender.Minute = DCFsyncMinute;

//                // Stunden
        //        Kalender.Stunde =    TimerBCD.Stunde & 0b00001111;
        //        Kalender.Stunde += ((TimerBCD.Stunde & 0b00110000) >> 4) * 10;
                Kalender.Stunde = DCFsyncStunde;

                // Tag
                Kalender.Wochentag = TimerBCD.Wochentag & 0b00000111;
                
#ifndef REDUCED_CODE_SIZE                
                Kalender.Tag =    TimerBCD.Tag & 0b00001111;
                Kalender.Tag += ((TimerBCD.Tag & 0b00110000) >> 3) * 10;

                // Monat
                Kalender.Monat =    TimerBCD.Monat & 0b00001111;
                Kalender.Monat += ((TimerBCD.Monat & 0b00010000) >> 3) * 10;
#endif
            }
            StoerungsfreieZeit = KEIN_DCF_MINUTEN;            // Fehleruhr zurcksetzen
        }

        DCFsyncMinuteAlt = DCFsyncMinute;
        DCFsyncStundeAlt = DCFsyncStunde;
    }
    
    DCFneueDaten = false;
 }


/******************************************************************************/
/* Interrupt Routines                                                         */
/******************************************************************************/
void interrupt intsrv(void)
{
    Disable_Watchdog(); // just in case we woke up from sleep by another interrupt - not the watchdog timeout
    
//    if (TMR0IF)
//    {
//        // Timer0 interrupt
//
//        TMR0IF = false;
//    }
    
    /* Timer1 ohne interrupt da in Verwendung fr CapSense Oszillator */
////    if (TMR1IF)
////    {
////        // Timer1 interrupt
////
////        TMR1IF = false;        // clear interrupt
////    }

//    else if (TMR2IF)
//    {
//        // Timer2 interrupt
//
//        TMR2IF = false;        // clear interrupt
//    }
    /**************************************************************************/
    if (TMR4IF)
    {
        // Timer4 interrupt (alle 10ms)
        
        // reset timer - not needed for Timer2/4/6 modules
//        PR4 = TIMER4_SYS;
//        TMR4 = 0;     
        
//        if (TimeOutOfSync == false)  // nicht ausfhren wenn Uhr verstellt wird
//        {
            ++SystemTicker;

            if ((SystemTicker % 2) == 0)
                SystemTickerZwanzigMs = true;
                
            // ------- DCF Timer - Minutensyncronisation
            DCFMinutensyncTimer--;
            if (DCFMinutensyncTimer == 0)
            {
                BitZeiger = 0;              // Alles auf Anfang
                Bitlaenge = 0;
                BitStart = false;           // Zeitmessung anhalten
            }

            // Zeitmessung fr die DCF Bitlnge
            if (BitStart && (Bitlaenge < 200))
                Bitlaenge++;
//        }
        
        
        if (((SystemTicker % 100) == 0) )//|| TimeOutOfSync)
        {
//            IO_RelayOne_Toggle();
            
            /* Uhr weiterstellen */
            Kalender.Sekunde++;
            if (Kalender.Sekunde >= 60)
            {
                Kalender.Sekunde = 0;
                Kalender.Minute++;
                
                if (StoerungsfreieZeit > 0)
                    StoerungsfreieZeit--;
            }
            
            if (Kalender.Minute >= 60)
            {
                Kalender.Minute = 0;
                Kalender.Stunde++;
            }
            
            if (Kalender.Stunde >= 24)
            {
                Kalender.Stunde = 0;

                /* Nchster Tag */
                Kalender.Wochentag ++;
                if (Kalender.Wochentag > 7)
                    Kalender.Wochentag = 1;
                
#ifndef REDUCED_CODE_SIZE
                Kalender.Tag ++;
                if (Kalender.Tag > (Monatslaenge[Kalender.Monat-1]))
                {
                    Kalender.Tag = 1;

                    /* Nchster Monat */
                    Kalender.Monat++;
                    if (Kalender.Monat > 12)
                    {
                        // year++
                        Kalender.Monat = 1;
                    }
                }
#endif
            }
//            if (TimeOutOfSync == false)  // nicht ausfhren wenn Uhr verstellt wird
//            {
                // decrement menu idle timer
                if (MenuIdleTimer > 0)
                    MenuIdleTimer--;
                else
                {
                    MenuLevel = 1;
                    MenuItem = 1;
                    MenuChangeValue = false;
                }
                if ((PowerOffTimer > 0) && !RFTX_Active)
                    PowerOffTimer--;

                if (DisplayON)  // do not update the display with new data, unless it is switched on
                    UpdateDisplay = true;
//            }
            
        } // Sekndlich (SystemTicker % 100)
        
        TMR4IF = false;        // clear interrupt
    }
    /**************************************************************************/
    else if (TMR6IF)    // Timer6 interrupt occured
    {        
        // Send data to RF module
        
        PR6 = RFTXTimerPR; // set new timer value, overwrite old PR6
        Timer6Config_HS();
        
        if (RFTXSendbyteActiveBit)  // set output pin accordingly
            IO_RFTxPort_SetHigh();
        else
            IO_RFTxPort_SetLow();
        
        RFTXTimerPR = 0;
        
        if (RFTXRepeatCounter == 0) // transmit complete, frame was send for n times
        {
            set_cpu_seed_low();
            RFTXRepeatCounter = RFTX_NUMBER_OF_SAMPLES;
            RFTX_Active = false;      // done - disable Timer6 interrupt
        }
        else
        {
            if (RFTXSendbyteActiveBitPos == 0)   // byte completed, get next one
            {
                RFTXSendbyteActiveBitPos = 7;
                ++RFTXSendBufferIndex;
            }
            
            if (RFTXSendBufferIndex == 16)  // end of buffer reached, full frame was send
            {
                RFTXSendBufferIndex = 0;
                --RFTXRepeatCounter;  // decrease re-transmit counter
            }
            else
            {
                RFTXSendbyte = RFTXSendBuffer[RFTXSendBufferIndex];     // get current byte from buffer
                do
                {
                    RFTXSendbyteActiveBit = (RFTXSendbyte >> RFTXSendbyteActiveBitPos) & 1; // check each bit, increase pulse length - stop if logical value changes
        //            PR6 = PR6 + 18;  // 18=288s, 36=576s, 54=864s - Pulse width
                    RFTXTimerPR += RFTX_Timer_Base; // increase timer for longer pulse/pause

                    if (RFTXSendbyteActiveBitPos == 0)
                        break;      // stop at last bit position
                    else
                        --RFTXSendbyteActiveBitPos;
                }
                while (((RFTXSendbyte >> RFTXSendbyteActiveBitPos) & 1) == RFTXSendbyteActiveBit); // compare next and current bit value
            }
        }
        
        TMR6IF = false;      // clear interrupt
    }
    /**************************************************************************/
    else if (IOCBF0)    // Interrupt-on-Change Flag bits - port RB0
    {
        // Minutensyncronisation zurcksetzen
        DCFMinutensyncTimer = 170; // 1.7 Sekunden herunterzhlen
        
        
        if (IOCBP0) // Steigende Flanke
        {
            // Nchster Interrupt auf fallender Flanke
            IOCBP0 = 0;
            IOCBN0 = 1;

            // Zeitmessung starten
            Bitlaenge = 0;
            BitStart = true;
            
        }
        else    // Fallende Flanke
        {
            // Nchster Interrupt wieder auf steigender Flanke
            IOCBP0 = 1;
            IOCBN0 = 0;
            
            // Zeitmessung beenden und Pulslnge auswerten
            BitStart = false;          // Zeitmessung anhalten
            
            if (Bitlaenge > 7)  // low = 100ms, high = 200ms
            {
                if (BitZeiger < 32)
                {
                    BitZeigerIndex = 0;
                    BitZeigerShift = BitZeiger;
                }
                else
                {
                    BitZeigerIndex = 1;
                    BitZeigerShift = BitZeiger -32;
                }
                        
                if (Bitlaenge > 15)     // Logische 1 detektiert, fragliches Bit auf 1 setzen
                {
//                    BitPuffer [(BitZeiger / 32)][BitPufferZeiger] |= ((unsigned long)1 << (BitZeiger-((BitZeiger / 32)*32)));
//                    BitPuffer [(BitZeiger / 32)] |= ((unsigned long)1 << (BitZeiger-((BitZeiger / 32)*32)));
                    BitPuffer[BitZeigerIndex] |= ((unsigned long)1 << BitZeigerShift);
                }
                else
                {
//                    BitPuffer [(BitZeiger / 32)][BitPufferZeiger] &= ~((unsigned long)1 << (BitZeiger-((BitZeiger / 32)*32)));
//                    BitPuffer [(BitZeiger / 32)] &= ~((unsigned long)1 << (BitZeiger-((BitZeiger / 32)*32)));
                    BitPuffer[BitZeigerIndex] &= ~((unsigned long)1 << BitZeigerShift);
                }

                // Nchstes Bit
                if (BitZeiger < 60)
                    BitZeiger++;
                
                if (BitZeiger == 59)    // In der letzten Sekunde erfolgt keine Absenkung (Minutensyncronisation)
                    DCFneueDaten = true;
            }
        }

        IOCBF0 = false;        // clear interrupt
    }

//    else if (IOCBF5)    // Interrupt-on-Change Flag bits - port RB5
//    {
//        IOCBF5 = false;        // clear interrupt
//    }
    
}
/* END Interrupt Routines                                                     */
/******************************************************************************/

/******************************************************************************/
/* Functions for the LCD menu                                                 */
/******************************************************************************/
// clear all EEPROM values for timer memory
void menu_exit_timer_clearall()
{
    unsigned char address = EE_ADDR_TIMER_START -1;
    
    do
    {
        address++;
        eeprom_write(address,0x00);
    }
    while (address < EE_ADDR_TIMER_END);
    
//    while(WR)
//        ;  // wait until last write is complete --- not needed, eeread will check and wait

    EEPROMUpdateFlag = true; // signal main loop to reload EEPROM values
    menu_exit_select_timer_memory();    // exit submenu and refresh
}

/* Write to EEPROM and exit submenu */
void menu_exit_eeprom_save_value()
{
    eeprom_write(Buff_NewValue_EEPROMAddr,*Buff_NewValue);
    
    EEPROMUpdateFlag = true; // signal main loop to reload EEPROM values
    menu_back();
}

/* Cyle through the days of a week and exit after last day */
void menu_exit_change_relay()
{
    if (MenuCurrentWeekday < 7)
        MenuCurrentWeekday++;     // move to next weekday on every OK button press
    else
        menu_exit_eeprom_save_value();
}

/* Enable or disable selected relay for current weekday */
void menu_relais_weekday_toggle()
{
    *Buff_NewValue ^= (1 << MenuCurrentWeekday);      // toggle bit of current day
}

void menu_enter_select_timer_memory()
{
//    Timer_Weekday_byte = eeprom_read(Timer_EEPROMAddr_Wday);
    Buff_NewValue = &Timer_Weekday_byte;    // get addres, write to this variable later via the pointer
    
    MenuChangeValue = true;
    MenuCurrentWeekday = 0;    
    menu_entersub();
}

/* Read the needed timer configuration bytes on exit of the "memory slot" selection submenu */
void menu_exit_select_timer_memory()
{
    // read corresponding hour and minute/relay values, used in other menu options
//    Timer_hour_byte = eeprom_read(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_H);
//    TimerM.byte = eeprom_read(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_M);
//    TimerR.byte = eeprom_read(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_R);
    
    menu_read_ee_timer();   // load global variables from eeprom
//    Timer_Minute = TimerM.bits.MINUTE;
    
    menu_back();
}

/* Read the needed timer configuration, when entering the menu to show correct values */
void menu_enter_timer()
{
    menu_read_ee_timer();   // load global variables from eeprom
    menu_entersub();
}

// load global variables from eeprom
void menu_read_ee_timer()
{
    // read corresponding hour and minute/relay values, used in other menu options
    Timer_hour_byte = eeprom_read(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_H);
    TimerM.byte = eeprom_read(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_M);
    TimerR.byte = eeprom_read(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_R);
    Timer_Minute = TimerM.bits.MINUTE;
}

/* Init and enter the submenu to change the weekday */
void menu_enter_change_timer_weekday()
{
    Buff_NewValue_EEPROMAddr = Timer_EEPROMAddr_Wday;   // save address to write to EEPROM later
    Buff_NewValue = &Timer_Weekday_byte;   // set pointer with address of current variable - will be written later
    
    MenuChangeValue = true;
    MenuCurrentWeekday = 1;
    menu_entersub();
}

void menu_enter_change_TimerRandomMax()
{
    Buff_NewValue_EEPROMAddr = EE_ADDR_TIMER_RND_MAX;   // save address to write to EEPROM later
    Buff_NewValue = &TimerRandomMinuteMax;   // set pointer with address of current variable - will be written later
    
    MenuChangeTimeHour = false; // display "minute", not hours
    MenuChangeValue = true;
    menu_entersub();
}

void menu_enter_TimerRandomMax_up()
{
    (*Buff_NewValue)++;
    if (*Buff_NewValue > 30)
        *Buff_NewValue = 0;
//    *Buff_NewValue &= 0b00011111;   // maximum value 31
}

void menu_enter_TimerRandomMax_down()
{
    (*Buff_NewValue)--;
    if (*Buff_NewValue == 0)
        *Buff_NewValue = 30;
//    *Buff_NewValue &= 0b00011111;   // minumum value >0?
}

/* Increase/decrease the Timer memory address to select a timer slot */
void menu_timer_addr_up()
{
    if (Timer_EEPROMAddr_Wday < EE_ADDR_TIMER_STARTa_H -1)
        Timer_EEPROMAddr_Wday++;
}
void menu_timer_addr_down()
{
    if (Timer_EEPROMAddr_Wday > EE_ADDR_TIMER_START)
        Timer_EEPROMAddr_Wday--;
}

/* Increase/decrease hour or minutes of the Timer */
void menu_timer_change_time_up()
{
    if (MenuChangeTimeHour)
    {
        if (Timer_hour_byte >= 23)
            Timer_hour_byte = 0;
        else
            Timer_hour_byte++;
    }
    else
    {
        if (Timer_Minute >= 59)
            Timer_Minute = 0;
        else
            Timer_Minute++;
    }
}
void menu_timer_change_time_down()
{
    if (MenuChangeTimeHour)
    {
        Timer_hour_byte--;
        if (Timer_hour_byte == 255)    // rolled over?
            Timer_hour_byte = 23;
    }
    else
    {
        Timer_Minute--;
        if (Timer_Minute == 255)    // rolled over?
            Timer_Minute = 59;
    }
}

/* Save hours and minutes, then exit the submenu */
void menu_exit_timer_change_time()
{
    if (MenuChangeTimeHour)
    {
        MenuChangeTimeHour = false;
    }
    else
    {
        TimerM.bits.MINUTE = Timer_Minute;
        menu_exit_timer_change_relay();
    }
}

/* Enters the submenu to change the timer or relay settings */
void menu_enter_timer_change_time_or_relay()
//void menu_enter_timer_change_time()
{
//    Timer_hour_byte = eeprom_read(Timer_EEPROMAddr_Wday+1);
//    TimerMR.byte = eeprom_read(Timer_EEPROMAddr_Wday+2);
// reading is done by exit function of memory slot selection function    
    
    MenuChangeTimeHour = true;
    MenuChangeValue = true;
    menu_entersub();
}

//void menu_enter_timer_change_relay()
//{
//    MenuChangeValue = true;
//    menu_entersub();
//}

/* Save and exit the time or relay settings menu */
void menu_exit_timer_change_relay()
{
    eeprom_write(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_H, Timer_hour_byte);
    eeprom_write(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_M, TimerM.byte);
    eeprom_write(Timer_EEPROMAddr_Wday+EE_ADDR_TIMER_STARTo_R, TimerR.byte);
    
    EEPROMUpdateFlag = true;

	menu_back();
}

/* Change the relay assigment/operation for the timer */
void menu_timer_change_relay_one()
{
    TimerR.bits.RelayOne++;    // we only care about the two lower bits
}
void menu_timer_change_relay_two()
{
    TimerR.bits.RelayTwo++;    // we only care about the two lower bits
}
void menu_timer_change_relay_three()
{
    TimerR.bits.RelayThree++;    // we only care about the two lower bits
}
void menu_timer_change_relay_four()
{
    TimerR.bits.RelayFour++;    // we only care about the two lower bits
}
void menu_timer_change_random_enabled()
{
    TimerM.bits.Random++;    // we only care about one bit
}
/* update the display */
void LCD_display()
{
    unsigned char DisplayZeile, AktuelleZeile;
    
    LCDClear();
    
    AktuelleZeile = MenuItem;

    for (DisplayZeile=0; DisplayZeile < 2; DisplayZeile++)
    {
        LCDGotoXY(0, DisplayZeile);
        
        if (AktuelleZeile == MenuItem && MenuItem > INFO_DISPLAY_LINES)     // zeilenmarkierung in displayzeile darstellen
        {
            if (MenuChangeValue)
                LCDChar(1); // custom char 1 - up/down
            else
//                LCDData(2); // custom char 2 - >arrow
                LCDChar(126); // char arrow
//            LCDGotoXY(1, DisplayZeile);
        }
        
        LCDWriteString(MainMenu[AktuelleZeile].text);
        MainMenu[AktuelleZeile].m_pointer();    // Menfunktion ausfhren
          

        if (menu_get_next_item() == AktuelleZeile)
            break;
        
        AktuelleZeile = menu_get_next_item();
    }

    UpdateDisplay = false;
}

/* Moves the curser/row up */
void menu_curup()
{
    MenuItem = menu_get_previous_item();
}

/* Moves the curser/row down */
void menu_curdown()
{
    MenuItem = menu_get_next_item();
}

/* Not called directly. Returns the index number of the next menu item of the same level */
unsigned char menu_get_next_item(void)
{
    unsigned char Index = MenuItem;
    
    do
    {
        Index++;
        if (MainMenu[Index].level == MenuLevel)
            return Index;
    }
    while (MainMenu[Index].level > MenuLevel);
    
    return MenuItem;
}

/* Not called directly. Returns the index number of the previous menu item of the same level */
unsigned char menu_get_previous_item(void)
{
    unsigned char Index = MenuItem;
    
    do
    {
        Index--;
        if (MainMenu[Index].level == MenuLevel)
            return Index;   
    }
    while (MainMenu[Index].level > MenuLevel);
    
    return MenuItem;
}

/* Step out from submenu */
void menu_back()
{
    MenuChangeValue = false;    // no "change value" cursor needed (is set on enter change settings function)
    MenuLevel--;    
    
    do
    {
        MenuItem--;
    }
    while (!(MainMenu[MenuItem].level == MenuLevel));
}

/* Enter next/sub menu level */
void menu_entersub()
{
    MenuItem++;
	MenuLevel = MainMenu[MenuItem].level;
}

/* does nothing */
void dummyfunc()
{
    //dummy
}

/* Read the just selected week from EEPROM and print it on the LCD */
void print_timer_weekday()
{
    Timer_Weekday_byte = eeprom_read(Timer_EEPROMAddr_Wday);
    Print_SelectWeekday();
}

/* Send text string to the LCD to show if hours or minutes are selected */
void print_timer_new_time_hhmm()
{
    if (MenuChangeTimeHour)
        LCDWriteString(menu_string_hour);
    else
        LCDWriteString(menu_string_minute);
}

/* Call to print new relay assignment for relay one */
void print_timer_relay_status_one()
{
    print_timer_status_relay(TimerR.bits.RelayOne);
}
/* Call to print new relay assignment for relay two */
void print_timer_relay_status_two()
{
    print_timer_status_relay(TimerR.bits.RelayTwo);
}
/* Call to print new relay assignment for relay three */
void print_timer_relay_status_three()
{
    print_timer_status_relay(TimerR.bits.RelayThree);
}
/* Call to print new relay assignment for relay four */
void print_timer_relay_status_four()
{
    print_timer_status_relay(TimerR.bits.RelayFour);
}

/* Print new relay action when setting relay assignment for the timer */
void print_timer_status_relay(unsigned char RelayToDisplay)
{
    LCDChar(':');
    
    if (RelayToDisplay & 0b00000010)
        Print_ON_OFF_Status(RelayToDisplay & 0b00000001);
    else
        LCDChar('-');
}

/* Print new time when setting on/off for the timer */
void print_timer_new_time()
{
    print_time(Timer_hour_byte, Timer_Minute);
}

void print_timer_random_enabled()
{
    LCDChar(':');
    Print_ON_OFF_Status(TimerM.bits.Random);
}

/* Shows char number value */
void print_changed_value()
{
    LCDWriteInt(*Buff_NewValue,3);
}

/* Info line with operational staus of relays */
void print_relais_status()
{
    LCDWriteString(LCD_RelayOne);
//    LCDData(':');
    Print_ON_OFF_Status(IORelayCurrentState & 0b00000001);
//    LCDData(' ');
    LCDWriteString(LCD_RelayTwo);
//    LCDData(':');
    Print_ON_OFF_Status(IORelayCurrentState & 0b00000010);
}
void print_relais_statusTwo()
{
    LCDWriteString(LCD_RelayThree);
    Print_ON_OFF_Status(IORelayCurrentState & 0b00000100);
//    LCDData(' ');
//    LCDWriteString(LCD_RelayFour);
//    Print_ON_OFF_Status(IORelayCurrentState & 0b00001000);
    
//    LCDWriteInt(TimerRandomMinute,3); // temp.... for debug info
}

/* Info line with time and DCF77 status */
void print_time_status()
{
//    LCDWriteInt(Kalender.Stunde,2);
//    LCDData(':');
//    LCDWriteInt(Kalender.Minute,2);
    print_time(Kalender.Stunde, Kalender.Minute);
//    LCDData(':');
//    LCDWriteInt(Kalender.Sekunde,2);
        
    LCDChar(' ');
    
    if (StoerungsfreieZeit > 0)
        LCDChar(0);      // LCD_DFC_OK
    else
        LCDChar('?');    // LCD_DCF_ERR
    
    if (LowBatt)
        LCDChar('!');    
}

/* Display time in hh:mm format */
void print_time(unsigned char Stunde, unsigned char Minute)
 {
    LCDWriteInt(Stunde,2);
    LCDChar(':');
    LCDWriteInt(Minute,2);
//    LCDData(' ');
 }

/* Prints "on" or "off" to the display */
void Print_ON_OFF_Status(char PortStatus)
{
    if (PortStatus)
        LCDWriteString(LCD_ON);
    else
        LCDWriteString(LCD_OFF);

    // same thing as above..
//    LCDWriteString(PortStatus ? LCD_ON : LCD_OFF);
}

void print_battery_status(void)
{
//    LCDWriteInt(ADCBatteryReading,4); // temp.... for debug info
//    LCDWriteInt(((ADCBatteryReading*4)/100),1); // temp.... for debug info
//    LCDChar(',');
//    LCDWriteInt(((ADCBatteryReading*4096)/10000),2); // temp.... for debug info
    
    if (LowBatt)
        LCDWriteString(menu_string_battery_low);
    else
        LCDWriteString(menu_string_battery_ok);
}

/* Prints 'x' or '-' for a day of the week,
 *  'xxxxxxx' or '-------' to represent fully selected/deselected week */
void Print_SelectWeekday()
{
    unsigned char i, UseChar;
    
    for (i=1; i < 8; i++)
    {        
        if ((*Buff_NewValue & (1 << i)))    // day is set
        {
            if (i == MenuCurrentWeekday)    // highlight selected
                UseChar = 248;
            else
                UseChar = 'x';
        }
        else
        {
            if (i == MenuCurrentWeekday)    // highlight selected
                UseChar = 253;
            else
                UseChar = '-';
        }
        LCDChar(UseChar);       // write to the LCD
    }
}

/* Prints currently selected EEPROM memory number */
void print_timer_memory_address()
{
    LCDWriteInt(Timer_EEPROMAddr_Wday -EE_ADDR_TIMER_START,2);
}

void print_debug_status(void)   // need to be removed when adding a 4th receiver...
{
//    LCDChar('a');
    LCDWriteInt(AnzahlAktuelleSchaltzeiten,2);
    LCDChar('w');
    LCDWriteInt(Kalender.Wochentag,1);
    LCDChar('s');
    LCDWriteInt(StoerungsfreieZeit,3);
}

/* EOF */