/******************************************************************************/
/* System Level #define Macros                                                */
/******************************************************************************/

/* Microcontroller MIPs (FCY) */
#define SYS_FREQ        400000L // normally running at 4MHz when using "delay" functions, temporary increase to 16MHz
#define FCY             SYS_FREQ/4
#define _XTAL_FREQ      SYS_FREQ

#define EEPROM_SIZE     255
#define __FLASHTYPE     1   //__FLASHTYPE == 1 defined in traditional devices that can write 1 word at a time eg. 16F877

//#define TIMER1_H    0xF6    // 10ms timer int @8MHz
//#define TIMER1_L    0x3B
//#define TIMER1_H    0xD8    // 10ms timer int @32MHz
//#define TIMER1_L    0xEF

//#define TIMER2  255  // Timer2 init

#define RFTX_Timer_Base     18
//#define Timer6Config()      T6CON = 0b00001111 // T6OUTPS = 0b0001, TMR6ON = 1, T6CKPS = 0b11 @ 32MHz
#define Timer6Config_HS()   T6CON = 0b00000111 // T6OUTPS = 0b0000, TMR6ON = 1, T6CKPS = 0b11 @ 16MHz
#define Timer6Config_LS()   T6CON = 0b00000110 // T6OUTPS = 0b0000, TMR6ON = 1, T6CKPS = 0b10 @ 4MHz
#define RFTX_Active         TMR6IE

/* 10 ms system timer */
#define TIMER4_SYS  155 // 156  // Timer4 init - for system ticker - 10ms timer
#define Timer4Config_HS()    T4CON = 0b00011111 // T4OUTPS = 0b0011, TMR4ON = 1, T4CKPS = 0b11 @ 16MHz
#define Timer4Config_LS()    T4CON = 0b00000111 // T4OUTPS = 0b0000, TMR4ON = 1, T4CKPS = 0b10 @ 4MHz

#define Enable_Watchdog()   WDTCON = 0b00000011 // activate with 2ms timeout
//#define Disable_Watchdog()  WDTCON = 0b00000000
#define Disable_Watchdog()  asm("bcf WDTCON,0") // must use "asm" single instruction to get prefetched before sleep

// switch from fast to slow CPU speed. Use for init as well
#define set_cpu_seed_low() \
    OSCCONbits.IRCF = 0b1101; \
    PR4 = TIMER4_SYS; \
    Timer4Config_LS(); \
    T6CONbits.TMR6ON = false   // turn off timer 6, RF TX module not used with slow CPU speed (4MHz)

// switch from slow to fast CPU speed
#define set_cpu_seed_high() \
    PR4 = TIMER4_SYS; \
    Timer4Config_HS(); \
    PR6 = RFTX_Timer_Base; \
    Timer6Config_HS(); \
    OSCCONbits.IRCF = 0b1111 // CPU @16 MHz

/* Pin assignment */
// DCF-77 module input - is using port interrupts!
//#define IO_DFC77_GetValue()         RB0


// Get random from LED input - using ADC on RB5
//#define IO_RelayOne_Toggle()           LATB5 ^= 1
#define ADC_Channel_Select_Rnd()     ADCON0bits.CHS = 0b00111   // channel AN7

// Battery voltage level - unsing ADC on RB1
#define ADC_Channel_Select_Batt()     ADCON0bits.CHS = 0b01011   // channel AN11

// Input Buttons (Voltage divider)
#define ADC_Channel_Select_Buttons()     ADCON0bits.CHS = 0b01000   // channel AN8

// DC-DC step-up module
#define IO_DCDC_Module_On()           LATB6 = 0
#define IO_DCDC_Module_Off()          LATB6 = 1
#define IO_DCDC_Module_GetValue()     RB6

// RF TX
#define IO_RFTxPort_GetValue()         RB7
#define IO_RFTxPort_Toggle()           LATB7 ^= 1
#define IO_RFTxPort_SetLow()           LATB7 = 0
#define IO_RFTxPort_SetHigh()          LATB7 = 1
#define IO_RFTxPort                    LATB7

/* EEPROM data addresses */
//// LDR - light sensor and input using 0x09 to 0x0e (0x0F free - blocked)
//#define EE_ADDR_LDR_IN       0x09    // first address used
//#define EE_ADDR_LDR_LEVEL    0x09    // Store threshold level (hysteresis is fixed)
//#define EE_Index_LDR_Level      0
////#define EE_ADDR_LDR_WDAY     0x0A    // Enable flag for each weekday, on & off for two relays = 4 EE addresses
//#define EE_ADDR_LDR_WDAY_ONE 0x0A
//#define EE_Index_LDR_WDay_One   1
//#define EE_ADDR_LDR_WDAY_TWO 0x0B
//#define EE_Index_LDR_WDay_Two   2
//#define EE_ADDR_LDR_SAMPLES  0x0C     // will result in average over n samples with n*10 seconds delay
//#define EE_Index_LDR_Samples    3
//#define EE_ADDR_IN_WDAY_ONE  0x0D
//#define EE_Index_IN_WDay_One    4
//#define EE_ADDR_IN_WDAY_TWO  0x0E
//#define EE_Index_IN_WDay_Two    5
//#define EE_COUNT_ADDR_LDR_IN    6    // using 6 consecutive EEPROM addresses

// address for max random value of timer minute
#define EE_ADDR_TIMER_RND_MAX   0x09

// address of timer values
/* use three bytes for one timer value: 1 byte for the week (actually using 7 bits), 
 * 1 byte for the hour, 1 byte for Minute (4 bits) + random enabled (1(2) bits), 1 byte relay assigment (4 bits) + operation (4 bits). */
#define EE_ADDR_TIMER_START     0x87
#define EE_ADDR_TIMER_END       0xFF
#define EE_ADDR_TIMER_STARTo_H   (EE_ADDR_TIMER_END - EE_ADDR_TIMER_START)/4     // address offset for hour values
#define EE_ADDR_TIMER_STARTa_H  EE_ADDR_TIMER_STARTo_H + EE_ADDR_TIMER_START     // start at address for hour value
#define EE_ADDR_TIMER_STARTo_M  ((EE_ADDR_TIMER_END - EE_ADDR_TIMER_START)/4)*2     // address offset for minute values
#define EE_ADDR_TIMER_STARTo_R  ((EE_ADDR_TIMER_END - EE_ADDR_TIMER_START)/4)*3     // address offset for realy values

/* EEPROM data addresses end */

// Bitmasken
//#define MaskeMinuten            0b00011111111000000000000000000000  // Inklusive Paritt
#define MaskeMinuten            0x1FE00000                          // Inklusive Paritt
#define ShiftMinuten            21                                  // Wieviele Bits nach Rechts ?

// #define MaskeStunden1           0b11100000000000000000000000000000
#define MaskeStunden1           0xE0000000
#define ShiftStunden1           29
//#define MaskeStunden2           0b00000000000000000000000000001111
#define MaskeStunden2           0xF
#define ShiftStunden2           3

// #define MaskeTag                0b00000000000000000000001111110000
#define MaskeTag                0x3F0
#define ShiftTag                4

// #define MaskeWochentag          0b00000000000000000001110000000000
#define MaskeWochentag                0x1C00
#define ShiftWochentag                10

// #define MaskeMonat              0b00000000000000111110000000000000
#define MaskeMonat              0x3E000
#define ShiftMonat              13

#define ON true
#define OFF false

// Wochentage (DCF77)
//#define MONTAG 1
//#define DIENSTAG 2
//#define MITTWOCH 3
//#define DONNERSTAG 4
//#define FREITAG 5
//#define SAMSTAG 6
//#define SONNTAG 7

/******************************************************************************/
/* System Function Prototypes                                                 */
/******************************************************************************/

/* Custom oscillator configuration funtions, reset source evaluation
functions, and other non-peripheral microcontroller initialization functions
go here. */

//void ConfigureOscillator(void); /* Handles clock switching/osc initialization */
//void set_cpu_seed_low(void);
//void set_cpu_seed_high(void);

//void Init_timer1(void);
//void Init_timer2(void);
void Init_timer4(void);
//void Init_timer6(void);
